@@ -1,4 +1,4 @@
-/* ZGV v1.3 - (c) 1993 Russell Marks for improbabledesigns.
+/* ZGV v2.0 - (c) 1993,1994 Russell Marks for improbabledesigns.
  * See README for license details.
  *
  * vgadisp.c - vga specific display routines.
@@ -13,26 +13,23 @@
 #include <signal.h>
 #include <sys/file.h>
 #include <vga.h>
+#include <vgagl.h>
 #include "gifeng.h"
 #include "usejpeg.h"
 #include "readnbkey.h"
 #include "handlevt.h"
 #include "helppage.h"
 #include "3deffects.h"
+#include "rc_config.h"
+#include "rcfile.h"
+#include "vgadisp.h"
 
 
-
-/* This 'default' is ONLY used by the 'mode sel' code when you switch from,
- * say, a 1-bit mono picture to a 256-colour one. (If you haven't got -
- * or svgalib doesn't support - a 640x480x256 mode on your card then it
- * falls back on 360x480x256. Let's not talk about laptops for now, ok? :)
- */
-#define DEFAULTMODE G640x480x256
-#define DEFAULTVIRT 0
-
-int curvgamode=DEFAULTMODE;
-int zoom=0,virtual=DEFAULTVIRT;
-int vkludge=0,modesel=0;
+/* zgv.c initialises these */
+int curvgamode;
+int zoom,virtual;
+int vkludge;
+int pixelsize;     /* in bytes - 1 for 8-bit, 3 for 24-bit */
 static int isjpeg;
 
 char viewerhelp[][80]={
@@ -70,12 +67,15 @@ struct {
 #define MOVSTP 10    /* for q,a,o,p panning */
 #define BIGSTP 100   /* for Q,A,O,P panning (i.e. with shift) */
 
+#define FIX_TO_EIGHT_BIT	1
+#define FIX_TO_HIGHCOLOUR	2
+
 byte *theimage;   /* quite slow if passed as parameter */
 int width,height,numcols; /* same here */
 
 double contrast=1.0;  /* note that double contrast is in fact 2 :) */
 int brightness=0;
-int scrnwide,scrnhigh,realscrnwide;
+int scrnwide,scrnhigh,realscrnwide,scrnpixelsize,scrncols;
 
 int palr[256],palg[256],palb[256];
 byte palr64[256],palg64[256],palb64[256];
@@ -84,7 +84,8 @@ int palrgb[768];
 int ttyfd;
 
 
-int readgiforjpg(char *,hffunc);
+#define GET15BITCOLOUR(r,g,b) (((r>>3)<<10)|((g>>3)<<5)|(b>>3))
+#define GET16BITCOLOUR(r,g,b) (((r>>3)<<11)|((g>>2)<<5)|(b>>3))
 
 
 int readgiforjpg(giffn,howfarfunc)
@@ -96,8 +97,17 @@ GIFINFO ginfo;
 byte *palette;
 
 /* at this stage, we know it ends in either .gif or .jpg */
+/* until I get around to adding PPM support, that is... :) */
 if(!strcmp(strrchr(giffn,'.'),".jpg"))
   {
+  /* don't *always* use 24-bit for JPEGs... check for 15/16/24-bit
+   * video modes before doing so.
+   */
+  pixelsize=has_highcolour_mode()?3:1;
+  if(pixelsize==3)
+    fix_to_similar_mode(FIX_TO_HIGHCOLOUR);
+  else
+    fix_to_similar_mode(FIX_TO_EIGHT_BIT);
   isjpeg=1;
   ginfo.numcols=256;
   theimage=NULL;
@@ -110,6 +120,11 @@ if(!strcmp(strrchr(giffn,'.'),".jpg"))
   }
 else
   {
+  /* part of the New Strategy (tm); always use 8-bit for GIFs.
+   * fix the mode to a similar 8-bit one if possible.
+   */
+  fix_to_similar_mode(FIX_TO_EIGHT_BIT);
+  pixelsize=1;
   isjpeg=0;
   result=readgif(giffn,&theimage,&palette,howfarfunc);
   if(result==_GIF_OK)
@@ -122,9 +137,6 @@ else
 if(((isjpeg)&&(result==1))||((!isjpeg)&&(result==_GIF_OK)))
   {
   numcols=ginfo.numcols;
-  if(((curvgamode<5)||(curvgamode==9))&&(ginfo.numcols>16))
-    { curvgamode=DEFAULTMODE; virtual=DEFAULTVIRT; }
-  if(!isjpeg) fixcurmode(&ginfo);
   showgif(palette);
   free(theimage);
   free(palette);
@@ -139,14 +151,98 @@ return(result);
 }
 
 
+/* modes aren't counted if they're locked out */
+int has_highcolour_mode()
+{
+int try;
+
+try=modematch(-1,-1,15);
+if(try!=-1)
+  if(cfg.mode_allowed[try])
+    return(1);
+    
+try=modematch(-1,-1,16);
+if(try!=-1)
+  if(cfg.mode_allowed[try])
+    return(1);
+    
+try=modematch(-1,-1,24);
+if(try!=-1)
+  if(cfg.mode_allowed[try])
+    return(1);
+
+return(0);
+}
+
+
+/* since we've already tested if it's possible, this can't fail */
+fix_to_similar_mode(modetype)
+int modetype;
+{
+int f,newmode,numpixels,pixdiff,newnp,newdiff,mode_ok;
+vga_modeinfo *vminfo;
+
+
+/* here's the plan;
+ * - we try to keep the number of pixels as close as possible:
+ *   as this will probably be quite noticeable, *this takes priority*.
+ * - we then try to match any 8-bit modes for EIGHT_BIT, or match
+ *   24/16/15-bit modes (in that order) for HIGHCOLOUR.
+ * - of course, the mode has to be either 8 or 15/16/24-bit as reqd.
+ */
+
+newmode=-1; pixdiff=(1<<30);
+numpixels=get_mode_width(curvgamode)*get_mode_height(curvgamode);
+
+/* we check 0 - 255 at the most */
+#ifdef HARM_HANEMAAYER_FIXES_THAT_SVGALIB_BUG
+for(f=0;f<256;f++)
+#else
+for(f=0;f<31;f++)
+#endif
+  {
+  if((!vga_hasmode(f))||(cfg.mode_allowed[f]==0)) continue;
+  vminfo=vga_getmodeinfo(f);
+  
+  mode_ok=0;
+  if(modetype==FIX_TO_EIGHT_BIT)
+    {
+    if(vminfo->colors==256) mode_ok=1;
+    }
+  else
+    {
+    if(vminfo->colors>256) mode_ok=1;
+    }
+  
+  if(mode_ok==0) continue;
+  
+  newnp=(vminfo->width)*(vminfo->height);
+  if((f==7)||(f==8)) newnp*=2;   /* account for effect of 'virtual' */
+  newdiff=numpixels-newnp;
+  if(newdiff<0) newdiff=(-newdiff);
+  
+  if(newdiff<=pixdiff)
+    {
+    newmode=f;
+    pixdiff=newdiff;
+    }
+  }
+
+/* we rely on the 15, 16, 24-bit mode number ordering to sort that
+ * out for us (hence the '<=' above).
+ */
+curvgamode=newmode;
+if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
+}
+
+
 is_this_file_jpeg()
 {
 return(isjpeg);
 }
 
 
-makerealpal(palette)
-byte *palette;
+makerealpal()
 {
 int f;
 
@@ -159,23 +255,6 @@ for(f=0;f<256;f++)
 }
 
 
-fixcurmode(ginfo)
-GIFINFO *ginfo;
-{
-if(!modesel) return;
-if(ginfo->numcols<=16)
-  {
-  curvgamode=4;
-  virtual=0;
-  }
-if((ginfo->width==320)&&(ginfo->height==200))
-  {
-  curvgamode=5;
-  virtual=0;
-  }
-}
-
-
 filterpal(palette)
 byte *palette;
 {
@@ -187,6 +266,7 @@ for(f=0;f<256;f++)   /* don't *really* need to know number of colours */
   palg[f]=dimmer(contrastup(palette[256+f]));
   palb[f]=dimmer(contrastup(palette[512+f]));
   }
+
 makerealpal();
 }
 
@@ -214,35 +294,83 @@ return((int)g);
 }
 
 
+int get_mode_width(vm)
+int vm;
+{
+vga_modeinfo *vmi;
+
+vmi=vga_getmodeinfo(vm);
+return(vmi->width);
+}
+
+int get_mode_height(vm)
+int vm;
+{
+vga_modeinfo *vmi;
+
+vmi=vga_getmodeinfo(vm);
+return(vmi->height);
+}
+
+int get_mode_pixelbytes(vm)
+int vm;
+{
+vga_modeinfo *vmi;
+
+vmi=vga_getmodeinfo(vm);
+return(vmi->bytesperpixel);
+}
+
+int get_mode_numcols(vm)
+int vm;
+{
+vga_modeinfo *vmi;
+
+vmi=vga_getmodeinfo(vm);
+return(vmi->colors);
+}
+
+
 graphicson()
 {
 int x;
-vga_modeinfo *vmi;
 
 /* workaround for svgalib bug */
-vga_setmode(TEXT);
-if(vga_hasmode(curvgamode))
+if((vga_hasmode(curvgamode))&&(cfg.mode_allowed[curvgamode]))
   vga_setmode(curvgamode);
 else
   {
-  /* we haven't? Aaargh!!! Ok, use 640x480 or 360x480 instead.
-   * This doesn't come under 'mode select' code, because we've got to
-   * do *something*...
-   */
-  if(vga_hasmode(G640x480x256))
+  /* we haven't? Aaargh!!! Ok, use 640x480 or 360x480 instead. */
+  if((vga_hasmode(G640x480x256))&&(cfg.mode_allowed[G640x480x256]))
     {
     vga_setmode(curvgamode=G640x480x256);
     virtual=0;
     }
   else
     {
-    vga_setmode(curvgamode=G360x480x256);
-    virtual=1;
+    if((vga_hasmode(G360x480x256))&&(cfg.mode_allowed[G360x480x256]))
+      {
+      vga_setmode(curvgamode=G360x480x256);
+      virtual=1;
+      }
+    else
+      {
+      /* *must* have 320x200 (see rcfile.c for more info) */
+      vga_setmode(curvgamode=G320x200x256);
+      virtual=0;
+      }
     }
   }
-vmi=vga_getmodeinfo(curvgamode);
-scrnwide=realscrnwide=vmi->width;
-scrnhigh=vmi->height;
+
+/* workaround for screen clearing probs in svgalib 0.95+ */
+vga_clear();
+if(pixelsize!=1) gl_setcontextvga(curvgamode);
+
+scrnwide=get_mode_width(curvgamode);
+scrnhigh=realscrnwide=get_mode_height(curvgamode);
+scrnpixelsize=get_mode_pixelbytes(curvgamode);
+scrncols=get_mode_numcols(curvgamode);
+
 if(virtual)
   scrnwide*=2;
 vga_setpalvec(0,256,palrgb);
@@ -258,7 +386,8 @@ if(vkludge)
 
 graphicsoff()
 {
-vga_setmode(TEXT);
+/* we let zgv.c deal with this now, so this is just a placeholder */
+/* vga_setmode(TEXT); */
 }
 
 
@@ -305,7 +434,10 @@ while(!quitshow)
   /* svga modes - note that F11=shift-F1 and F12=shift-F2 */
   if(((key>=RK_F1)&&(key<=RK_F12))||(key>=RK_SHIFT_F3)&&(key<=RK_SHIFT_F10))
     /* the RK_F?'s are defined such that this sort of thing is ok to do: */
-    if(vga_hasmode(key-RK_F1+10))
+    /* (here comes the 'if' statement from hell...) */
+    if((vga_hasmode(key-RK_F1+10))&&(cfg.mode_allowed[key-RK_F1+10])&&
+       (((pixelsize==1)&&(get_mode_numcols(key-RK_F1+10)==256))||
+        ((pixelsize==3)&&(get_mode_numcols(key-RK_F1+10)>256))))
       {
       curvgamode=key-RK_F1+10;   /* so F1 = mode 10 */
       virtual=0;
@@ -316,14 +448,23 @@ while(!quitshow)
   if((key>='0')&&(key<='9'))
     {
     if(key=='0')
+      {
       virtual=(virtual==1)?0:1;
+      graphicson();
+      redraw=1;
+      }
     else
       {
-      curvgamode=key-48;
-      if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
+      if((vga_hasmode(key-48))&&(cfg.mode_allowed[key-48])&&
+         (((pixelsize==1)&&(get_mode_numcols(key-48)==256))||
+          ((pixelsize==3)&&(get_mode_numcols(key-48)>256))))
+        {
+        curvgamode=key-48;
+        if((curvgamode==7)||(curvgamode==8)) virtual=1; else virtual=0;
+        graphicson();
+        redraw=1;
+        }
       }
-    graphicson();
-    redraw=1;
     }
   else
     switch(key)
@@ -352,11 +493,14 @@ while(!quitshow)
       case 'r': fx_rot();    px=py=0; redraw=1; break;
       case 'z':
         zoom=(!zoom); redraw=1; px=py=0; graphicson(); break;
+#if 0
+/* this appears to be a bit broken with VC switching after ^Z */
       case 26:   /* ^Z */
         vga_setmode(TEXT);
         raise(SIGTSTP);
         graphicson(); redraw=1;
         break;
+#endif
       case '?':
         showhelp(ttyfd,"- KEYS FOR FILE DISPLAY SCREEN -",viewerhelp);
         /* falls through to 'refresh screen' */
@@ -391,7 +535,7 @@ redrawgif(px,py)
 int px,py;
 {
 int x,y,xdim;
-byte *realline;
+byte *realline,*ptr;
 long place;
 
 if(zoom)
@@ -401,40 +545,68 @@ else
   if(width-px<scrnwide) xdim=width-px; else xdim=scrnwide;
   if((py>=height)||(px>=width)) return(0);
   /* hopefully the following is fairly quick as fewer ifs... ? */
-  if(virtual)
+  if((virtual)&&(pixelsize==1))
     {
     if((realline=malloc(scrnwide))==NULL) return(0);
-    if(xdim==scrnwide)
-      {for(y=0;(y<height-py)&&(y<scrnhigh);y++)
-        {for(x=0;(x<width-px)&&(x<scrnwide);x++)
-          *(realline+(x>>1))=getvpix(px,py,x,y);
-        vga_drawscanline(y,realline);}}
-    else
-      {for(y=0;(y<height-py)&&(y<scrnhigh);y++)
-        {for(x=0;x<xdim;x++)
-          *(realline+(x>>1))=getvpix(px,py,x,y);
-        vga_drawscansegment(realline,0,y,xdim>>1);}}
+    for(y=0;(y<height-py)&&(y<scrnhigh);y++)
+      {
+      for(x=0;x<xdim;x++)
+        *(realline+(x>>1))=getvpix(px,py,x,y);
+        vga_drawscansegment(realline,0,y,xdim>>1);
+      }
     free(realline);
     }
   else
     {
-    if(xdim==scrnwide)
-      {for(y=0;(y<height-py)&&(y<scrnhigh);y++)
-        vga_drawscanline(y,theimage+(py+y)*width+px);}
-    else
-      {for(y=0;(y<height-py)&&(y<scrnhigh);y++)
-        vga_drawscansegment(theimage+(py+y)*width+px,0,y,xdim);}
+    switch(scrnpixelsize)
+      {
+      case 3:
+        gl_putboxpart(0,0,(scrnwide<width)?scrnwide:width,
+	        (scrnhigh<height)?scrnhigh:height,width,height,theimage,px,py);
+        break;
+      
+      default:
+        for(y=0;(y<height-py)&&(y<scrnhigh);y++)
+          eventuallydrawscansegment(theimage+pixelsize*((py+y)*width+px),
+          	0,y,xdim*pixelsize);
+        break;
+      }
     }
   }
 }
 
 
+eventuallydrawscansegment(ptr,x,y,len)
+byte *ptr;
+int x,y,len;
+{
+unsigned short buf[2048];
+
+if(scrnpixelsize!=2)
+  vga_drawscansegment(ptr,x,y,len);
+else
+  {
+  int f,a;
+  unsigned short *bufref;
+
+  if(scrncols==32768)
+    for(f=0,bufref=buf;f<len;f+=3,bufref++)
+      *bufref=GET15BITCOLOUR(ptr[f+2],ptr[f+1],ptr[f]);
+  else
+    for(f=0,bufref=buf;f<len;f+=3,bufref++)
+      *bufref=GET16BITCOLOUR(ptr[f+2],ptr[f+1],ptr[f]);
+
+  vga_drawscansegment((byte *)buf,x,y,(len/3)<<1);
+  }
+}
+
+
 int getvpix(px,py,x,y)
 int px,py,x,y;
 {
 int p1,p2,r,g,b;
 
-if(vkludge)
+if((vkludge)&&(pixelsize==1))
   {
   p1=*(theimage+(py+y)*width+px+x);
   if(px+x+1>=width) return(p1);
@@ -445,7 +617,7 @@ if(vkludge)
   return(closest(r,g,b));
   }
 else
-  return(*(theimage+(py+y)*width+px+x));
+  return(*(theimage+pixelsize*((py+y)*width+px+x)));
 }
 
 
@@ -455,14 +627,14 @@ drawzoomedgif()
 register int a,b,x,yp,yw;
 long y,sw,sh,lastyp;
 int xoff,yoff;
-int c,pixwide,pixhigh,pr,pg,pb;
+int c,cr,cg,cb,pixwide,pixhigh,pr,pg,pb;
 long place;
 int bigimage;
 byte *rline;
 int tmp1,tmp2,tr,tg,tb,tn;
 
-if((rline=malloc(scrnwide))==NULL) return(0);
-for(x=0;x<scrnwide;x++)
+if((rline=malloc(scrnwide*pixelsize))==NULL) return(0);
+for(x=0;x<scrnwide*pixelsize;x++)
   rline[x]=0;
 
 /* try landscapey */
@@ -481,8 +653,8 @@ if(bigimage)
   graphicson();
   pixhigh=(int)(((float)height)/((float)sh)+0.5);
   pixwide=(int)(((float)width)/((float)sw)+0.5);
-  /*if(pixhigh==0)*/ pixhigh++;
-  /*if(pixwide==0)*/ pixwide++;
+  pixhigh++;
+  pixwide++;
   for(y=0;y<height;y++)
     {
     yp=(y*sh)/height;
@@ -497,20 +669,42 @@ if(bigimage)
           for(b=0;(b<pixhigh)&&(y+b<height);b++)
             for(a=0;(a<pixwide)&&(x+a<width);a++)
               {
-              tmp2=*(theimage+yw+a+b*width);
-              tr+=palr64[tmp2];
-              tg+=palg64[tmp2];
-              tb+=palb64[tmp2];
+              if(scrncols<=256)
+                {
+                tmp2=*(theimage+yw+a+b*width);
+                tr+=palr64[tmp2];
+                tg+=palg64[tmp2];
+                tb+=palb64[tmp2];
+                }
+              else
+                {
+                tb+=*(theimage+3*(yw+a+b*width));
+                tg+=*(theimage+3*(yw+a+b*width)+1);
+                tr+=*(theimage+3*(yw+a+b*width)+2);
+                }  
               tn++;
               }
           tr/=tn; tg/=tn; tb/=tn;
-          rline[(x*sw)/width]=closest(tr,tg,tb);
+          if(scrncols<=256)
+            rline[(x*sw)/width]=closest(tr,tg,tb);
+          else
+            {
+            tmp1=(x*sw)/width*pixelsize;
+            rline[tmp1]=tb;
+            rline[tmp1+1]=tg;
+            rline[tmp1+2]=tr;
+            }
           }
         }
       else
         for(x=0;x<width;x++,yw++)
-          rline[(x*sw)/width]=*(theimage+yw);
-      vga_drawscanline(yp,rline);  
+          {
+          if(scrncols<=256)
+            rline[(x*sw)/width]=*(theimage+yw);
+          else
+            memcpy(rline+(x*sw)/width*3,theimage+yw*3,3);
+          }
+      eventuallydrawscansegment(rline,0,yp,scrnwide*pixelsize);
       lastyp=yp;
       }
     }
@@ -525,52 +719,60 @@ else
 
   free(rline);  /* get rid of that... */
     /* and get a screen's worth... hahahahahahahahaha er, sorry. */
-  if((rline=malloc(scrnwide*(scrnhigh+pixhigh)))==NULL) return(0);
+  if((rline=malloc(pixelsize*scrnwide*(scrnhigh+pixhigh)))==NULL) return(0);
     /* notice that we allow for a screwy 'pixhigh' value to avoid
        any nasty old segfaults! oh, now we wipe it: */
-  for(x=0;x<scrnwide*scrnhigh;x++) rline[x]=0;
+  for(x=0;x<scrnwide*scrnhigh*pixelsize;x++) rline[x]=0;
   
   if(virtual)
     sw>>=1;   /* leave pixwide though as it *might* miss with >> */
   for(y=0;y<height;y++)
     {
     yoff=(y*sh)/height;
-    yw=y*width;
-    for(x=0;x<width;x++,yw++)
+    yw=y*width*pixelsize;
+    for(x=0;x<width;x++,yw+=pixelsize)
       {
-      c=*(theimage+yw);
+      if(scrncols<=256)
+        c=*(theimage+yw);
+      else
+        {
+        cb=*(theimage+yw); cg=*(theimage+yw+1); cr=*(theimage+yw+2);
+        }
       xoff=(x*sw)/width;
       for(b=0;b<pixhigh;b++)
         for(a=0;a<pixwide;a++)
-          *(rline+(yoff+b)*scrnwide+xoff+a)=c;
+          if(scrncols<=256)
+            *(rline+(yoff+b)*scrnwide+xoff+a)=c;
+          else
+            {
+            tmp1=3*((yoff+b)*scrnwide+xoff+a);
+            *(rline+tmp1)=cb;
+            *(rline+tmp1+1)=cg;
+            *(rline+tmp1+2)=cr;
+            }
       }
-    vga_drawscanline(y,rline+y*scrnwide); /* really pessimistic */
+    eventuallydrawscansegment(rline+pixelsize*y*scrnwide,
+    	0,y,scrnwide*pixelsize); /* really pessimistic */
     }
   for(;y<scrnhigh;y++)          /* and do the rest */
-    vga_drawscanline(y,rline+y*scrnwide);
+    eventuallydrawscansegment(rline+pixelsize*y*scrnwide,
+    				0,y,scrnwide*pixelsize);
   free(rline);
   }
 }
 
 
-putrgbpixel(x,y,c)
-int x,y,c;
-{
-vga_setcolor(c);
-vga_drawpixel(x,y);
-}
-
-
 fx_mirror()
 {
 byte *tmp;
 int x,y;
 
-tmp=malloc(width*height);
+tmp=malloc(width*height*pixelsize);
 if(tmp==NULL) return(0);
 for(y=0;y<height;y++)
   for(x=0;x<width;x++)
-    *(tmp+y*width+(width-x-1))=*(theimage+y*width+x);
+    memcpy(tmp+pixelsize*(y*width+(width-x-1)),
+    	theimage+pixelsize*(y*width+x),pixelsize);
 
 free(theimage);
 theimage=tmp;
@@ -582,11 +784,11 @@ fx_flip()
 byte *tmp;
 int x,y;
 
-tmp=malloc(width*height);
+tmp=malloc(width*height*pixelsize);
 if(tmp==NULL) return(0);
 for(y=0;y<height;y++)
-  for(x=0;x<width;x++)
-    *(tmp+(height-y-1)*width+x)=*(theimage+y*width+x);
+  memcpy(tmp+pixelsize*((height-y-1)*width),
+  	theimage+pixelsize*(y*width),width*pixelsize);
 
 free(theimage);
 theimage=tmp;
@@ -598,11 +800,12 @@ fx_rot()
 byte *tmp;
 int x,y;
 
-tmp=malloc(width*height);
+tmp=malloc(width*height*pixelsize);
 if(tmp==NULL) return(0);
 for(y=0;y<height;y++)
   for(x=0;x<width;x++)
-    *(tmp+(height-y-1)+x*height)=*(theimage+y*width+x);
+    memcpy(tmp+pixelsize*((height-y-1)+x*height),
+    	theimage+pixelsize*(y*width+x),pixelsize);
 x=height;y=width;
 width=x;height=y;
 
